"""
This file defines the entities needed by our legislation.

Taxes and benefits can be calculated for different entities: persons, household, companies, etc.

See https://openfisca.org/doc/key-concepts/person,_entities,_role.html
"""
from typing import Any

from policyengine_core.entities import build_entity

# TODO: Consider removing use of setattr so that we can use mypy for Entities
Household: Any = build_entity(
    key="household",
    plural="households",
    label="All the people in a family or group who live together in the same place.",
    doc="""
    Household is an example of a group entity.
    A group entity contains one or more individual·s.
    Each individual in a group entity has a role (e.g. parent or children). Some roles can only be held by a limited number of individuals (e.g. a 'first_parent' can only be held by one individual), while others can have an unlimited number of individuals (e.g. 'children').

    Example:
    Housing variables (e.g. housing_tax') are usually defined for a group entity such as 'Household'.

    Usage:
    Check the number of individuals of a specific role (e.g. check if there is a 'second_parent' with household.nb_persons(Household.SECOND_PARENT)).
    Calculate a variable applied to each individual of the group entity (e.g. calculate the 'salary' of each member of the 'Household' with salaries = household.members("salary", period = MONTH); sum_salaries = household.sum(salaries)).

    For more information, see: https://openfisca.org/doc/coding-the-legislation/50_entities.html
    """,
    roles=[
        {
            "key": "member",
            "plural": "members",
            "label": "Member",
            "doc": "A member of the household.",
        },
    ],
)

Person = build_entity(
    key="person",
    plural="people",
    label="An individual. The minimal legal entity on which a legislation might be applied.",
    doc="""

    Variables like 'salary' and 'income_tax' are usually defined for the entity 'Person'.

    Usage:
    Calculate a variable applied to a 'Person' (e.g. access the 'salary' of a specific month with person("salary", "2017-05")).
    Check the role of a 'Person' in a group entity (e.g. check if a the 'Person' is a 'first_parent' in a 'Household' entity with person.has_role(Household.FIRST_PARENT)).

    For more information, see: https://openfisca.org/doc/coding-the-legislation/50_entities.html
    """,
    is_person=True,
)

entities = [Household, Person]



from policyengine_ng.model_api import *


class is_tax_exempt(Variable):
    label = "is tax exempt"
    entity = Person
    definition_period = YEAR
    value_type = bool

    def formula(person, period, parameters):
        employment_income = person("employment_income", period)
        p = parameters(period).gov.tax
        employment_income_at_or_below_threshold = (
            employment_income <= p.exempt_threshold
        )
        gross_income = person("gross_income", period)
        has_non_employment_income = gross_income > employment_income
        return (
            employment_income_at_or_below_threshold
            & ~has_non_employment_income
        )



from policyengine_ng import (
    Simulation,
    CountryTaxBenefitSystem,
    system as tax_benefit_system,
)
from policyengine_core.parameters import get_parameter
from policyengine_core.periods import instant
from policyengine_core.enums import Enum
import dpath
import json
import logging


def calculate(household: dict, reform: dict) -> dict:
    if len(reform.keys()) > 0:
        system = tax_benefit_system.clone()
        for parameter_name in reform:
            for time_period, value in reform[parameter_name].items():
                start_instant, end_instant = time_period.split(".")
                parameter = get_parameter(system.parameters, parameter_name)
                node_type = type(parameter.values_list[-1].value)
                try:
                    value = float(value)
                except:
                    pass
                parameter.update(
                    start=instant(start_instant),
                    stop=instant(end_instant),
                    value=node_type(value),
                )
    else:
        system = tax_benefit_system

    simulation = Simulation(
        tax_benefit_system=system,
        situation=household,
    )

    household = json.loads(json.dumps(household))

    requested_computations = get_requested_computations(household)

    for (
        entity_plural,
        entity_id,
        variable_name,
        period,
    ) in requested_computations:
        try:
            variable = system.get_variable(variable_name)
            result = simulation.calculate(variable_name, period)
            population = simulation.get_population(entity_plural)
            if "axes" in household:
                count_entities = len(household[entity_plural])
                entity_index = 0
                for _entity_id in household[entity_plural].keys():
                    if _entity_id == entity_id:
                        break
                    entity_index += 1
                result = (
                    result.astype(float)
                    .reshape((-1, count_entities))
                    .T[entity_index]
                    .tolist()
                )
                # If the result contains infinities, throw an error
                if any([math.isinf(value) for value in result]):
                    raise ValueError("Infinite value")
                else:
                    household[entity_plural][entity_id][variable_name][
                        period
                    ] = result
            else:
                entity_index = population.get_index(entity_id)
                if variable.value_type == Enum:
                    entity_result = result.decode()[entity_index].name
                elif variable.value_type == float:
                    entity_result = float(str(result[entity_index]))
                    # Convert infinities to JSON infinities
                    if entity_result == float("inf"):
                        entity_result = "Infinity"
                    elif entity_result == float("-inf"):
                        entity_result = "-Infinity"
                elif variable.value_type == str:
                    entity_result = str(result[entity_index])
                else:
                    entity_result = result.tolist()[entity_index]

                household[entity_plural][entity_id][variable_name][
                    period
                ] = entity_result
        except Exception as e:
            if "axes" in household:
                pass
            else:
                logging.warn(f"Error computing {variable_name}: {e}")
                household[entity_plural][entity_id][variable_name][
                    period
                ] = None

    return household


def get_requested_computations(household: dict):
    requested_computations = dpath.util.search(
        household,
        "*/*/*/*",
        afilter=lambda t: t is None,
        yielded=True,
    )
    requested_computation_data = []

    for computation in requested_computations:
        path = computation[0]
        entity_plural, entity_id, variable_name, period = path.split("/")
        requested_computation_data.append(
            (entity_plural, entity_id, variable_name, period)
        )

    return requested_computation_data



from policyengine_ng.model_api import *


class household_net_income(Variable):
    label = "net income"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-NGN"
    adds = ["household_market_income", "household_benefits"]
    subtracts = ["household_tax"]



from policyengine_ng.model_api import *


class tax(Variable):
    label = "individual tax"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-NGN"

    def formula(person, period, parameters):
        exempt = person("is_tax_exempt", period)
        main_rates = person("main_tax_rates", period)
        minimum_tax = person("minimum_tax", period)
        return ~exempt * max_(main_rates, minimum_tax)



from policyengine_ng.model_api import *


class employment_income(Variable):
    label = "employment income"
    documentation = "Total employment income, including bonuses and tips."
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-NGN"



from policyengine_ng.model_api import *


class consolidated_relief_allowance(Variable):
    label = "consolidated relief allowance"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-NGN"

    def formula(person, period, parameters):
        gross_income = person("gross_income", period)
        p = parameters(period).gov.tax.consolidated_relief_allowance
        return p.flat + (gross_income * p.percent)



from policyengine_ng.model_api import *


class age(Variable):
    label = "age"
    documentation = "Age of the person in years."
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "year"
    default_value = 18



from policyengine_ng.model_api import *


class Gender(Enum):
    MALE = "Male"
    FEMALE = "Female"


label = "Gender"


class gender(Variable):
    label = "gender"
    entity = Person
    definition_period = YEAR
    value_type = Enum
    possible_values = Gender
    default_value = Gender.MALE



from policyengine_ng.model_api import *


class market_income(Variable):
    label = "market income"
    documentation = "Total non-government income."
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-NGN"
    adds = ["employment_income"]



import logging
import os
from pathlib import Path
from typing import Dict, Type

from policyengine_ng import entities
from policyengine_core.country_template.data.datasets.country_template_dataset import (
    CountryTemplateDataset,
)
from policyengine_core.data.dataset import Dataset
from policyengine_core.populations.population import Population
from policyengine_core.simulations import (
    Microsimulation as CoreMicrosimulation,
)
from policyengine_core.simulations import Simulation as CoreSimulation
from policyengine_core.taxbenefitsystems import TaxBenefitSystem

from .constants import COUNTRY_DIR

DATASETS = [CountryTemplateDataset]  # Important: must be instantiated


class CountryTaxBenefitSystem(TaxBenefitSystem):
    entities = entities.entities
    variables_dir = COUNTRY_DIR / "variables"
    parameters_dir = COUNTRY_DIR / "parameters"
    auto_carry_over_input_variables = False
    modelled_policies = COUNTRY_DIR / "modelled_policies.yaml"
    basic_inputs = [
        "age",
        "employment_income",
    ]


system = CountryTaxBenefitSystem()


class Simulation(CoreSimulation):
    default_tax_benefit_system = CountryTaxBenefitSystem
    default_tax_benefit_system_instance = system
    default_role = "parent"
    default_calculation_period = "2022-01"
    default_input_period = "2022-01"


class Microsimulation(CoreMicrosimulation):
    default_tax_benefit_system = CountryTaxBenefitSystem
    default_dataset = CountryTemplateDataset
    default_tax_benefit_system_instance = system
    default_dataset_year = 2022
    default_calculation_period = "2022-01"



from policyengine_ng.model_api import *


class marginal_tax_rate(Variable):
    label = "marginal tax rate"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "/1"

    def formula(person, period, parameters):
        mtr_values = np.zeros(person.count, dtype=np.float32)
        simulation = person.simulation
        adult_index_values = person("adult_index", period)
        DELTA = 1_000
        mtr_adult_count = parameters(
            period
        ).simulation.marginal_tax_rate_adults
        print(simulation.input_variables)
        for adult_index in range(1, 1 + mtr_adult_count):
            alt_simulation = simulation.get_branch(
                f"adult_{adult_index}_pay_rise"
            )
            mask = adult_index_values == adult_index
            for variable in simulation.tax_benefit_system.variables:
                if variable not in simulation.input_variables:
                    alt_simulation.delete_arrays(variable)
            alt_simulation.set_input(
                "employment_income",
                period,
                person("employment_income", period) + mask * DELTA,
            )
            alt_person = alt_simulation.person
            household_net_income = person.household(
                "household_net_income", period
            )
            household_net_income_higher_earnings = alt_person.household(
                "household_net_income", period
            )
            increase = (
                household_net_income_higher_earnings - household_net_income
            )
            mtr_values += where(mask, 1 - increase / DELTA, 0)
            print(household_net_income)
            print(household_net_income_higher_earnings)
            print(increase)
            print(mtr_values)
        return mtr_values


class adult_index(Variable):
    value_type = int
    entity = Person
    label = "Index of adult in household"
    definition_period = YEAR

    def formula(person, period, parameters):
        return (
            person.get_rank(
                person.household,
                -person("age", period),
                condition=person("is_adult", period),
            )
            + 1
        )



import streamlit as st

st.set_page_config(layout="wide")

hide_footer_style = """
<style>
header {
    display: none !important;
}
footer {
    display: none !important;
}
section > div.block-container {
    padding-top: 0px !important;
    padding-bottom: 0px !important;
}
h1,
h2,
h3,
h4,
h5,
h6,
p,
span,
div {
  font-family: "Roboto", sans-serif !important;
  font-weight: 500;
}
[data-baseweb="slider"] {
    padding-left: 10px !important;
}
#MainMenu {
    visibility: hidden;
}
footer {
    visibility: hidden;
}
.modebar{
      display: none !important;
}
</style>
"""
st.write(hide_footer_style, unsafe_allow_html=True)

from streamlit_ace import st_ace
import json
from api_utils import calculate

default_situation = {
    "household": {
        "people": {
            "parent": {
                "employment_income": {2023: 30_000},
                "tax": {2023: None},
            },
        },
        "households": {
            "household": {
                "members": ["parent"],
            },
        },
    },
    "policy": {},
}
data_input, api_output = st.tabs(["JSON input", "API output"])
with data_input:
    st.caption("Describe people, households and reforms")
    situation = json.loads(
        st_ace(
            language="json",
            theme="github",
            value=json.dumps(default_situation, indent=4),
            height=300,
        )
    )
    try:
        result = calculate(situation["household"], situation["policy"])
    except Exception as e:
        st.error(e)

with api_output:
    st.caption("PolicyEngine's API computes their taxes and benefits")
    st.json(result)



from policyengine_ng.model_api import *


class minimum_tax(Variable):
    label = "minimum tax"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-NGN"

    def formula(person, period, parameters):
        gross_income = person("gross_income", period)
        p = parameters(period).gov.tax
        return p.minimum * gross_income



from policyengine_ng.model_api import *


class household_benefits(Variable):
    label = "benefits"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-NGN"



from policyengine_ng.model_api import *


class taxable_income(Variable):
    label = "taxable income"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-NGN"

    def formula(person, period, parameters):
        gross_income = person("gross_income", period)
        consolidated_relief_allowance = person(
            "consolidated_relief_allowance", period
        )
        return max_(0, gross_income - consolidated_relief_allowance)



import argparse
import os


def main():
    # First, find the current package version number from the setup.py file
    with open("setup.py", "r") as f:
        setup = f.read()
    version = setup.split("version=")[1].split(",")[0].strip('"')
    # Then, clone the https://github.com/policyengine/policyengine-api repo using the GitHub CLI
    pat = os.environ["GITHUB_TOKEN"]
    os.system(
        f"git clone https://nikhilwoodruff:{pat}@github.com/policyengine/policyengine-api"
    )
    # Then, cd inside and run gcp/bump_country_package.py --country policyengine-uk --version {version}
    os.system(
        f"cd policyengine-api && python gcp/bump_country_package.py --country policyengine-ng --version {version}"
    )


if __name__ == "__main__":
    main()



from pathlib import Path

from setuptools import find_packages, setup

# Read the contents of our README file for PyPi
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

# Please make sure to cap all dependency versions, in order to avoid unwanted
# functional and integration breaks caused by external code updates.

general_requirements = [
    "policyengine-core",
    "numpy",
]

dev_requirements = [
    "black",
    "coverage",
    "jupyter-book",
    "plotly",
    "pytest",
    "setuptools",
    "wheel",
    "yaml-changelog>=0.1.7",
    "linecheck",
    "furo<2023",
    "markupsafe==2.0.1",
    "sphinx>=4.5.0,<5",
    "sphinx-argparse>=0.3.2,<1",
    "sphinx-math-dollar>=1.2.1,<2",
]

setup(
    name="policyengine-ng",
    version="0.5.1",
    author="PolicyEngine",
    author_email="hello@policyengine.org",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    description="Core microsimulation engine enabling country-specific policy models.",
    keywords="tax benefit microsimulation framework",
    license="https://www.fsf.org/licensing/licenses/agpl-3.0.html",
    license_files=("LICENSE",),
    url="https://github.com/policyengine/policyengine-core",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": [
            "policyengine-core=policyengine_core.scripts.policyengine_command:main",
        ],
    },
    extras_require={
        "dev": dev_requirements,
    },
    include_package_data=True,  # Will read MANIFEST.in
    install_requires=general_requirements,
    packages=find_packages(exclude=["tests*"]),
)



from policyengine_ng.model_api import *


class is_adult(Variable):
    label = "is an adult"
    documentation = "Whether this individual is over 18."
    entity = Person
    definition_period = YEAR
    value_type = bool

    def formula(person, period, parameters):
        age = person("age", period)
        return age >= 18



from policyengine_ng.model_api import *


class main_tax_rates(Variable):
    label = "main tax rates"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-NGN"

    def formula(person, period, parameters):
        taxable_income = person("taxable_income", period)
        p = parameters(period).gov.tax
        return p.main_rates.calc(taxable_income)
